"""
The :mod:`skmine.itemset` module includes methods which consider reccurent itemsets.
An itemset is an unsorted set of items, eg. identifiers for books in a store, words in a text, etc.
"""
from .lcm import LCM
from .slim import SLIM
