"""
The :mod:`skmine.graph` module includes methods which consider recurrent subgraphs called pattern in a data graph .
A graph is a structure that represents data as a set of entities called "nodes", connected by relationships
called "edges". In a labeled graph, nodes and edges can have associated labels that allow to represent attributes of entities and relationships.
"""