"""
scikit-mine
Pattern mining algorithms for the Python ecosystem
"""

# TODO : define logger
# see : https://github.com/joblib/joblib/issues/634#issuecomment-367321025
# joblib currently disable logging inside delayed functions
__author__ = """scikit-mine team"""
__email__ = """remi.adon@gmail.com"""
__version__ = "0.0.8"  # pylint: disable= undefined-variable, pointless-statement
__credits__ = "Centre Inria Rennes - Bretagne Atlantique"
